#! /usr/bin/env python
# -*- coding: utf-8 -*-
#
import sys
import os
import femagtools.mcv as mcv
import logging

logger = logging.getLogger(__name__)


class Reader(object):
    def __init__(self, filename):
        self.mc1_type = mcv.MAGCRV

        with open(filename) as f:
            self.name = os.path.splitext(
                os.path.basename(f.readline().strip()))[0]
            r = f.readline().split()
            self.angle = [float(a) for a in r[1:]]
            self.curve = [dict(hi=[], bi=[])
                          for i in range(int(r[0])-1)]
            for l in f:
                r = l.split()
                if len(r) > 1:
                    bh = [float(x.replace(',', '.')) for x in r]
                    if bh[1] > 0:
                        b = bh[0]
                        for i, h in enumerate(bh[1:]):
                            self.curve[i]['hi'].append(h)
                            self.curve[i]['bi'].append(b)
            if len(self.curve) > 1:
                self.mc1_type = mcv.ORIENT_CRV
        logger.info("JHB %s curves %d sizes %s", self.name, len(self.curve),
                    [len(c['bi']) for c in self.curve])
        
    def __getitem__(self, index):
        return self.__getattribute__(index)
    
    def getValues(self):
        """return values as mcv dict"""
        return {
            'name': self.name,
            'ctype': self.mc1_type,
            'curve': self.curve}


def read(filename):
    """read JHB File and return mc dict"""
    jhb = Reader(filename)
    return jhb.getValues()


if __name__ == "__main__":
    if len(sys.argv) == 2:
        filename = sys.argv[1]
    else:
        filename = sys.stdin.readline().strip()
            
    jhb = Reader(filename)
    print(jhb.getValues())



# -*- coding: utf-8 -*-
"""
    femagtools.femag
    ~~~~~~~~~~~~~~~~

    Running FEMAG


"""
import subprocess
import os
import glob
import logging
try:
    import zmq
except ImportError:
    pass
import sys
import json
import io
import femagtools.mcv
import femagtools.airgap as ag
import femagtools.fsl
import femagtools.ntib as ntib
import femagtools.config as cfg
import time
import re
import  threading

logger = logging.getLogger(__name__)


BCHEXT = '.BATCH' if sys.platform.startswith('linux') else '.BCH'  # win32

def get_shortCircuit_parameters(bch, nload):
    try:
        if nload < 0: nload=0
        if nload > 2: nload=2
        if nload > 0:
            dqld = bch.dqPar['ld']
            dqlq = bch.dqPar['lq']
            dqpsim = bch.dqPar['psim']
            if len(dqld) <= nload or len(dqlq) <= nload or len (dqpsim) <= nload:
                ld = dqld[-1]/bch.armatureLength
                lq = dqlq[-1]/bch.armatureLength
                psim = dqpsim[-1]/bch.armatureLength
            else:
                ld = dqld[nload-1]/bch.armatureLength
                lq = dqlq[nload-1]/bch.armatureLength
                psim = dqpsim[nload-1]/bch.armatureLength
        else:
            ld = bch.machine['ld']/bch.armatureLength
            lq = bch.machine['lq']/bch.armatureLength
            psim = bch.machine['psim']/bch.armatureLength
        return dict(
                r1 = bch.machine['r1'],
                ld = ld,
                lq = lq,
                psim = psim,
                num_pol_pair = bch.machine['p'],
                fc_radius = bch.machine['fc_radius'],
                lfe = bch.armatureLength/1e3,
                pocfilename = bch.machine['pocfile'],
                num_par_wdgs = bch.machine['num_par_wdgs'],
                calculationMode = 'shortcircuit')
    except (KeyError, AttributeError, IndexError):
        raise FemagError("missing pm/Rel-Sim results")
    
class FemagError(Exception):
    pass


class BaseFemag(object):
    def __init__(self, workdir, cmd, magnetizingCurves, magnets):
        self.workdir = workdir
        if cmd:
            self.cmd = cmd
        else:
            self.cmd = cfg.get_femag()

        if magnetizingCurves:
            if isinstance(magnetizingCurves,
                          femagtools.mcv.MagnetizingCurve):
                self.magnetizingCurves = magnetizingCurves
            else:
                self.magnetizingCurves = femagtools.mcv.MagnetizingCurve(
                    magnetizingCurves)
        else:
            self.magnetizingCurves = []
        if magnets:
            if isinstance(magnets, femagtools.Magnet):
                self.magnets = magnets
            else:
                self.magnets = femagtools.Magnet(magnets)
        else:
            self.magnets = []

    def copy_magnetizing_curves(self, model, dir=None):
        """extract mc names from model and write files into workdir or dir if given

        Return:
            list of extracted mc names (:obj:`list`)
        """
        dest = dir if dir else self.workdir
        return [self.magnetizingCurves.writefile(m[0], dest, fillfac=m[1])
                for m in model.set_magcurves(
                        self.magnetizingCurves, self.magnets)]

    def create_fsl(self, pmMachine, simulation):
        """create list of fsl commands"""
        self.model = femagtools.MachineModel(pmMachine)
        self.modelname = self.model.name
        self.copy_magnetizing_curves(self.model)

        builder = femagtools.fsl.Builder()
        if simulation:
            return builder.create(self.model, simulation, self.magnets)
        return builder.create_model(self.model, self.magnets) + ['save_model("cont")']
        

    def get_log_value(self, pattern, modelname='FEMAG-FSL.log'):
        result = []
        pat = re.compile(pattern)
        with open(os.path.join(self.workdir, 'FEMAG-FSL.log')) as f:
            for line in f:
                m = pat.search(line)
                if m:
                    result.append(float(line.split(':')[-1].split()[0]))
        return result

    def get_bch_file(self, modelname, offset=0):
        """return latest bch file (if any)"""
        bchfile_list = sorted(glob.glob(os.path.join(
            self.workdir, modelname+'_[0-9][0-9][0-9]'+BCHEXT)))
        if(bchfile_list):
            return bchfile_list[-1-offset]
        return ''
    
    def read_bch(self, modelname=None, offset=0):
        "read most recent BCH/BATCH file and return result"
        # read latest bch file if any
        if not modelname:
            modelname = self._get_modelname_from_log()

        result = femagtools.bch.Reader()
        bchfile = self.get_bch_file(modelname, offset)
        if bchfile:
            logger.info("Read BCH {}".format(bchfile))
            with io.open(bchfile, encoding='latin1',
                         errors='ignore') as f:
                result.read(f)
        return result

    def read_isa(self, modelname=None):
        "read most recent I7/ISA7 file and return result"
        if not modelname:
            modelname = self._get_modelname_from_log()
        import femagtools.isa7
        return femagtools.isa7.read(os.path.join(self.workdir, modelname))

    def read_los(self, modelname=None):
        "read most recent LOS file and return result"
        # read latest los file if any
        if not modelname:
            modelname = self._get_modelname_from_log()

        losfile_list = sorted(glob.glob(os.path.join(
            self.workdir, modelname+'_[0-9][0-9][0-9].LOS')))
        if len(losfile_list) > 0:
            return ntib.read_los(losfile_list[-1])

        return dict()

    def read_airgap_induction(self, modelname='', offset=0):
        """read airgap induction"""
        # we need to figure out the number of poles in model
        bch = self.read_bch(modelname, offset)
        return ag.read(os.path.join(self.workdir, 'bag.dat'),
                       bch.machine['p_sim'])

    def _get_modelname_from_log(self):
        """
        Read the modelname from the Femag Log file
        """
        with open(os.path.join(self.workdir, 'FEMAG-FSL.log')) as f:
            for l in f:
                if l.startswith('New model') or l.startswith('Load model'):
                    model = l.split('"')[1]
                    break
        return model


class Femag(BaseFemag):
    """Invoke and control execution of FEMAG

    Args:
        workdir: name of working directory
        cmd: name of femag program (default wfemag64 on windows, xfemag64 on linux)
        magnetizingCurves: collection of lamination material curves
        magnets: collection of magnet material
    """
    def __init__(self, workdir, cmd=None,
                 magnetizingCurves=None, magnets=None):
        super(self.__class__, self).__init__(workdir, cmd,
                                             magnetizingCurves, magnets)

    def run(self, filename, options=['-b'], fsl_args=[]):
        """invoke FEMAG in current workdir

        Args:
            filename: name of file to execute
            options: list of FEMAG options
            fsl_args: list of FSL argument options

        Raises:
            FemagError
        """
        if self.cmd.startswith('wfemag') and \
           '-b' in options and \
           '-m' not in options:
            options.insert(0, '-m')
        args = [self.cmd] + options + [filename] + fsl_args

        basename, ext = os.path.splitext(os.path.basename(filename))
        outname = os.path.join(self.workdir, basename+'.out')
        errname = os.path.join(self.workdir, basename+'.err')
        with open(outname, 'w') as out, open(errname, 'w') as err:
            logger.info('invoking %s', ' '.join(args))
            proc = subprocess.Popen(
                args,
                stdout=out, stderr=err, cwd=self.workdir)

        proc.wait()
        errs = []
        # print femag output
        with io.open(outname, encoding='latin1', errors='ignore') as outfile:
            for l in outfile:
                if l.find('ERROR') > -1:
                    errs.append(l.strip())

        rc = proc.returncode
        logger.info("%s exited with returncode %d (num errs=%d)",
                    self.cmd, rc, len(errs))
        if rc != 0 or errs:
            with io.open(errname, encoding='latin1',
                         errors='ignore') as errfile:
                for l in errfile:
                    errs.append(l.strip())
            errs.insert(0, 'Exit code {}'.format(rc))
            raise FemagError(errs)
        
    def cleanup(self):
        "removes all created files in workdir"
        if not os.path.exists(self.workdir):
            return
        cleanfiles = ('*.B*CH', '*.I*7-*', '*.A*7-*', '*.nc-*',
                      '*.dat', '*.LOS', '*.svg', '*.png', '*.hxy')
        # '*.TMC','*.TMO', '*.PROT'):
        for p in cleanfiles:
            for f in glob.glob(os.path.join(self.workdir, p)):
                os.remove(f)

    def __call__(self, pmMachine, simulation={},
                 options=['-b'], fsl_args=[]):
        """setup fsl file, run calculation and return 
        BCH or LOS results if any."""
        fslfile = 'femag.fsl'
        with open(os.path.join(self.workdir, fslfile), 'w') as f:
            f.write('\n'.join(self.create_fsl(pmMachine,
                                              simulation)))
        if simulation:
            if 'poc' in simulation:
                 with open(os.path.join(self.workdir,
                                        simulation['pocfilename']), 'w') as f:
                    f.write('\n'.join(simulation['poc'].content()))
            if simulation['calculationMode'] == "pm_sym_loss":
                with open(os.path.join(self.workdir,
                                       self.modelname+'.ntib'), 'w') as f:
                    f.write('\n'.join(ntib.create(
                        simulation['speed'],
                        simulation['current'],
                        simulation['angl_i_up'])))
                # TODO: add r1, m
 
        self.run(fslfile, options, fsl_args)
        if simulation:
            if simulation['calculationMode'] == "pm_sym_loss":
                return self.read_los(self.modelname)

            bch = self.read_bch(self.modelname)
            if simulation['calculationMode'] == 'pm_sym_fast':
                if simulation.get('shortCircuit', False):
                    logger.info("short circuit simulation")
                    simulation.update(
                        get_shortCircuit_parameters(bch,
                                                    simulation.get('initial',2)))
                    builder = femagtools.fsl.Builder()
                    fslcmds = (builder.open_model(self.model) +
                               builder.create_shortcircuit(simulation))
                    with open(os.path.join(self.workdir, fslfile), 'w') as f:
                        f.write('\n'.join(fslcmds))
                    self.run(fslfile, options)
                    bchfile = self.get_bch_file(self.modelname)
                    if bchfile:
                        bchsc = femagtools.bch.Reader()
                        logger.info("Read BCH {}".format(bchfile))
                        with io.open(bchfile, encoding='latin1',
                                     errors='ignore') as f:
                            bchsc.read(f)
                    bch.scData = bchsc.scData
                    for w in bch.flux:
                        try:
                            bch.flux[w] += bchsc.flux[w]
                            bch.flux_fft[w] += bchsc.flux_fft[w]
                        except (KeyError, IndexError):
                            logging.debug("No additional flux data in sc simulation")
                            break

                    bch.torque += bchsc.torque
                    bch.demag += bchsc.demag
            return bch
        return dict(status='ok', message=self.modelname)


class FemagTask(threading.Thread):
    def __init__(self, port, args, workdir, logdir):
        threading.Thread.__init__ (self)
        self.port = port
        self.args = args + [str(self.port)]
        self.returncode = None
        self.workdir = workdir
        self.logdir = logdir
        
    def run(self):
        logger.info("femag is ready on port %d workdir %s",
                    self.port, self.workdir)
        outname = os.path.join(self.logdir, f'femag-{self.port}.out')
        errname = os.path.join(self.logdir, f'femag-{self.port}.err')
        with open(outname, 'w') as out, open(errname, 'w') as err:
            self.proc = subprocess.Popen(
                self.args,
                stdout=out, stderr=err, cwd=self.workdir)
        
        self.returncode = self.proc.wait()


class SubscriberTask(threading.Thread):
    def __init__(self, port, host, notify):
        threading.Thread.__init__ (self)
        context = zmq.Context.instance()
        self.subscriber = context.socket(zmq.SUB)
        if not host:
            host = 'localhost'
        self.subscriber.connect(f'tcp://{host}:{port}')
        self.subscriber.setsockopt(zmq.SUBSCRIBE, b'')
        self.controller = zmq.Context.instance().socket(zmq.PULL)
        self.controller_url = 'inproc://publisher'
        self.controller.bind(self.controller_url)
        self.poller = zmq.Poller()
        self.poller.register(self.subscriber, zmq.POLLIN)
        self.poller.register(self.controller, zmq.POLLIN)
        self.logger = logger
        self.notify = notify
    
    def stop(self):
        socket = zmq.Context.instance().socket(zmq.PUSH)
        socket.connect(self.controller_url)
        socket.send(b"quit")
        socket.close()
        
    def run(self):
        self.logger.info("subscriber is ready")
        while True:
            socks = dict(self.poller.poll())
            if socks.get(self.subscriber) == zmq.POLLIN:
                try:
                    response = self.subscriber.recv_multipart()
                    # Sometimes femag send messages with only len = 1. These messages must be ignored
                    if len(response) < 2:
                        continue
                    self.notify([s.decode('latin1') for s in response])

                except Exception:
                    self.logger.error("error in subscription messag processing", exc_info=True)

            if socks.get(self.controller) == zmq.POLLIN:
                req = self.controller.recv()
                self.logger.info(req)
                break
            
        self.logger.debug("subscriber stopped")


class ZmqFemag(BaseFemag):
    """Invoke and control execution of FEMAG with ZeroMQ

    Args:
        port: port number of req socket
        host: hostname (ip addr)
        workdir: name of working directory
        logdir: name of logging directory (default is workdir/log)
        cmd: name of femag program
    """
    def __init__(self, port, host='localhost', workdir='', logdir='',
                 cmd=None,
                 magnetizingCurves=None, magnets=None):
        super(self.__class__, self).__init__(workdir, cmd,
                                             magnetizingCurves, magnets)
        self.host = host
        self.port = port
        self.pubhost = ''
        self.logdir = logdir if logdir else os.path.join(workdir,'log')
        if not os.path.exists(self.logdir):
            os.makedirs(self.logdir)

        self.request_socket = self.__req_socket()
        self.subscriber = None

    def close(self):
        if self.subscriber:
            self.subscriber.stop()
            
        if self.request_socket:
            logger.debug("close request_socket")
            self.request_socket.close()

        logger.debug("done")

    def __del__(self):
        self.close()
        logger.debug("Destructor ZmqFemag")

    def __req_socket(self):
        """returns a new request client"""
        context = zmq.Context.instance()
        self.request_socket = context.socket(zmq.REQ)
        url = 'tcp://{0}:{1}'.format(
            self.host, self.port)
        logger.debug("connect %s", url)
        self.request_socket.connect(url)
        return self.request_socket

    def subscribe(self, notify):
        """attaches a notify function"""
        logger.info("Subscribe on '%s'", self.pubhost)
        if self.subscriber is None:
            self.subscriber = SubscriberTask(self.port+1, self.pubhost, notify)
            self.subscriber.start()
        else:
            # reattach?
            self.subscriber.notify = notify
        return 

    def __is_running(self):
        try:
            return self.femagTask.returncode is None
        except:
            pass
        return False
         
    def send_request(self, msg, pub_consumer=None, timeout=None):
        if timeout:
            self.request_socket.setsockopt(zmq.RCVTIMEO, timeout)
            self.request_socket.setsockopt(zmq.LINGER, 0)
        else:
            self.request_socket.setsockopt(zmq.RCVTIMEO, -1)  # default value
            self.request_socket.setsockopt(zmq.LINGER, 30000)  # default value

        while True:
            try:
                for m in msg[:-1]:
                    self.request_socket.send_string(m, flags=zmq.SNDMORE)
                if isinstance(msg[-1], list):
                    self.request_socket.send_string('\n'.join(msg[-1]))
                else:
                    self.request_socket.send_string(msg[-1])
                logger.debug("msg %s", msg[-1])
                return self.request_socket.recv_multipart()
            except zmq.error.Again:
                pass
                
            except Exception as e:
                #logger.exception("send_request")
                logger.info("send_request: %s Message %s", str(e), msg)
                return [b'{"status":"error", "message":"' + str(e).encode() + b'"}']

    def send_fsl(self, fsl, timeout=None):
        """sends FSL commands in ZMQ mode and blocks until commands are processed

        Args:
            fsl: string (or list) of FSL commands
            timeout: The maximum time (in milliseconds) to wait for a response

        Return:
            status
        """
        header = 'FSL'
        self.request_socket.send_string(header, flags=zmq.SNDMORE)

        if isinstance(fsl, list):
            self.request_socket.send_string('\n'.join(fsl))
        else:
            self.request_socket.send_string(fsl)
        logger.debug("Sent fsl wait for response")

        if timeout:
            self.request_socket.setsockopt(zmq.RCVTIMEO, timeout)
            self.request_socket.setsockopt(zmq.LINGER, 0)
        else:
            self.request_socket.setsockopt(zmq.RCVTIMEO, -1)  # default value
            self.request_socket.setsockopt(zmq.LINGER, 30000)  # default value
        import datetime
        startTime = datetime.datetime.now() if timeout else None
        while True:
            try:
                response = self.request_socket.recv_multipart()
                return [s.decode('latin1') for s in response]  # NOTE: femag encoding is old school
            except zmq.error.Again as e:
                logger.info("Again [%s], timeout: %d", str(e), timeout)
                if not startTime:
                    continue

                diffTime = datetime.datetime.now() - startTime
                logger.debug("Diff msec[%d]", diffTime.microseconds)
                if diffTime.microseconds < timeout:
                    continue

                logger.info("ALERT not running close socket")
                self.close()
                return ['{"status":"error", "message":"Femag is not running"}', '{}']

            except Exception as e:
                logger.exception("send_fsl")
                logger.error("send_fsl: %s", str(e))
                if timeout:  # only first call raises zmq.error.Again
                    return ['{"status":"error", "message":"Femag is not running"}', '{}']
                msg = str(e)
                return ['{"status":"error", "message":"'+msg+'"}', '{}']

    def run(self, options=['-b'], restart=False, procId=None):  # noqa: C901
        """invokes FEMAG in current workdir and returns pid

        Args:
            options: list of FEMAG options
        """
        if self.__is_running():
            if restart:
                logger.info("must restart")
                self.quit(True)
                self.femagTask.join()
                logger.info("Stopped procId: %s", self.femagTask.proc.pid)
            else:
                return self.femagTask.proc.pid

        if self.cmd.startswith('wfemag') and \
           '-b' in options and \
           '-m' not in options:
            options.insert(0, '-m')

        args = [self.cmd] + options
        self.femagTask = FemagTask(self.port, args, self.workdir, self.logdir)
        self.femagTask.start()

        # check if mq is ready for listening
        lcount = 10
        for t in range(lcount):
            time.sleep(0.1)
            if self.__is_running():
                logger.info("femag (pid: '{}') is listening".format(
                    self.femagTask.proc.pid))
                break
            
        return self.femagTask.proc.pid

    def quit(self, save_model=False):
        """terminates femag"""

        if not self.__is_running():
            logger.info("Femag already stopped")
            return

        if self.subscriber:
            self.subscriber.stop()
            self.subscriber = None
            
        # send exit flags
        f = '\n'.join(['exit_on_end = true',
                       'exit_on_error = true'])
        response = self.send_fsl(f)

        # send quit command
        try:
            response = [r.decode('latin1')
                        for r in self.send_request(
                                ['CONTROL', 'quit'], timeout=10000)]
        except Exception as e:
            logger.error("Femag Quit zmq message %s", e)

        logger.debug("Sent QUIT to femag %s", response)
        # if query, send a answer
        obj = json.loads(response[0])
        logger.debug("status: {}".format(obj['status']))
        if obj['status'] == 'Query':
            logger.info('query: %s => %s',
                        obj['message'], 'saved' if save_model else 'not saved')

            # Only send one msg
            response = self.request_socket.send_string(
                'Ok' if save_model else 'Cancel')
        
    def upload(self, files):
        """upload file or files
        returns number of transferred bytes
        (FEMAG 8.5 Rev 3282 or greater only)
        """
        ret = []
        total = 0
        chunk_size = 20*1024
        fnames = files
        if isinstance(files, str):
            fnames = [files]
        for fn in fnames:
            self.request_socket.send_string('CONTROL', flags=zmq.SNDMORE)
            basename = os.path.basename(fn)
            logger.info("upload %s --> %s", fn, basename)
            self.request_socket.send_string('upload = {}'.format(basename),
                                            flags=zmq.SNDMORE)
            with open(fn, mode="rb") as file:
                while True:
                    data = file.read(chunk_size)
                    if not data:
                        break
                    more = 0 if len(data) < chunk_size else zmq.SNDMORE
                    self.request_socket.send(data, flags=more)
                    total += len(data)
            ret += [s.decode('latin1')
                    for s in self.request_socket.recv_multipart()]
        return ret
    
    def copyfile(self, filename, dirname):
        """copy filename to dirname (FEMAG 9.2)"""
        request_socket = self.__req_socket()
        request_socket.send_string('CONTROL', flags=zmq.SNDMORE)
        request_socket.send_string(f'copyfile {filename} {dirname}')
        return [r.decode() for r in request_socket.recv_multipart()]
        
    def change_case(self, dirname):
        """change case to dirname (FEMAG 9.2)"""
        request_socket = self.__req_socket()
        request_socket.send_string('CONTROL', flags=zmq.SNDMORE)
        request_socket.send_string(f'casedir {dirname}')
        return [r.decode() for r in request_socket.recv_multipart()]
        
    def delete_case(self, dirname):
        """delete case dir (FEMAG 9.2)"""
        request_socket = self.__req_socket()
        request_socket.send_string('CONTROL', flags=zmq.SNDMORE)
        request_socket.send_string(f'casedir -{dirname}')
        return [r.decode() for r in request_socket.recv_multipart()]
        
    def cleanup(self, timeout=2000):
        """remove all FEMAG files in working directory 
        (FEMAG 8.5 Rev 3282 or greater only)"""
        return [r.decode('latin1')
                for r in self.send_request(['CONTROL', 'cleanup'], timeout=timeout)]
    
    def release(self):
        """signal finish calculation task to load balancer to free resources
        (Docker Cloud environment only)
        """
        return [r.decode('latin1')
                for r in self.send_request(['close'])]

    def info(self, timeout=2000):
        """get various resource information 
        (FEMAG 8.5 Rev 3282 or greater only)"""
        response = [r.decode('latin1')
                    for r in self.send_request(['CONTROL', 'info'],
                                               timeout=timeout)]
        try:
            self.pubhost = json.loads(response[1])['addr']
            logger.info("Set pubhost %s", self.pubhost)
        except KeyError:
            # ignore pubhost setting if no addr is returned (Windows only)
            pass
        return response

    def publishLevel(self, level):
        """set publish level"""
        return [r.decode('latin1')
                for r in self.send_request(['CONTROL', 'publish = {}'
                                            .format(level)], timeout=10000)]

    def getfile(self, filename=''):
        """get file (FEMAG 8.5 Rev 3282 or greater only)"""
        response = self.send_request(
            ['CONTROL', 'getfile = {}'.format(filename)])
        return [response[0].decode('latin1'),
                response[1] if len(response) else b'']
        
    def exportsvg(self, fslcmds):
        """get svg format from fsl commands (if any graphic created)
        (since FEMAG 8.5 Rev 3343) """
        response = self.send_request(['SVG', fslcmds])
        rc = json.loads(response[0].decode('latin1'))
        if rc['status'] == 'ok':
            return self.getfile(rc['result_file'][0])
        return [s.decode('latin1') for s in response]

    def interrupt(self):
        """send push message to control port to stop current calculation"""
        context = zmq.Context.instance()
        if not self.pubhost:
            self.pubhost = '127.0.0.1'
        ctrl = context.socket(zmq.PUSH)
        ctrl.connect('tcp://{0}:{1}'.format(
                self.pubhost, self.port+2))
        logger.info("Interrupt %s", self.pubhost)
        ctrl.send_string('interrupt')
        ctrl.close()

    def copy_magnetizing_curves(self, model, dir=None):
        """extract mc names from model and write files into workdir or dir if given
           and upload to Femag

        Return:
            list of extracted mc names (:obj:`list`)
        """
        dest = dir if dir else self.workdir
        for m in model.set_magcurves(
                self.magnetizingCurves, self.magnets):
            f = self.magnetizingCurves.writefile(m[0], dest, fillfac=m[1])
            self.upload(os.path.join(dest, f))
    
    def __call__(self, pmMachine, simulation):
        """setup fsl file, run calculation and return BCH results
        Args:
          pmMachine: dict with machine parameters or name of model
          simulation; dict with simulation parameters

        Raises:
           FemagError
        """
        if isinstance(pmMachine, str):
            modelpars = dict(name=pmMachine)
        else:
            modelpars = pmMachine
        if 'exit_on_end' not in modelpars:
            modelpars['exit_on_end'] = 'false'
        if 'exit_on_error' not in modelpars:
            modelpars['exit_on_error'] = 'false'
        response = self.send_fsl(['save_model("close")'] +
                                 self.create_fsl(modelpars,
                                                 simulation))
        r = json.loads(response[0])
        if r['status'] != 'ok':
            raise FemagError(r['message'])

        result_file = r['result_file'][0]
        if simulation['calculationMode'] == "pm_sym_loss":
            return self.read_los(self.modelname)

        status, content = self.getfile(result_file)
        r = json.loads(status)
        if r['status'] == 'ok':
            bch = femagtools.bch.Reader()
            bch.read(content.decode('latin1'))
            if simulation['calculationMode'] == 'pm_sym_fast':
                if simulation.get('shortCircuit', False):
                    logger.info("Short Circuit")
                    simulation.update(
                        get_shortCircuit_parameters(bch,
                                                    simulation.get('initial', 2)))
                    builder = femagtools.fsl.Builder()
                    response = self.send_fsl(builder.create_shortcircuit(simulation))
                    r = json.loads(response[0])
                    if r['status'] != 'ok':
                        raise FemagError(r['message'])
                    result_file = r['result_file'][0]
                    status, content = self.getfile(result_file)
                    r = json.loads(status)
                    if r['status'] == 'ok':
                        bch.read(content.decode('latin1'))
                    
            return bch
        raise FemagError(r['message'])



# -*- coding: utf-8 -*-
"""
    femagtools.grid
    ~~~~~~~~~~~~~~~

    Parameter range calculation



"""
import logging
import glob
import os
import time
import numpy as np
import femagtools
import femagtools.model
import femagtools.fsl
import femagtools.condor
import femagtools.moproblem
import femagtools.getset
import shutil
import functools

logger = logging.getLogger(__name__)


def get_report(decision_vars, objective_vars, objectives, domain):
    """returns a combined list of objective and design values
    with header"""
    x = create_parameter_range(domain)
    y = np.reshape(np.asarray(objectives),
                   (np.shape(objectives)[0], np.shape(x)[0])).T
    c = np.arange(x.shape[0]).reshape(x.shape[0], 1)
    return [[d['label'] for d in decision_vars] +
            [o['label']
             for o in objective_vars] + ['Directory'],
            [d['name'] for d in decision_vars] +
            [o['name']
             for o in objective_vars]] + [
                     z[:-1].tolist() + [int(z[-1])]
                     for z in np.hstack((x, y, c))]


def baskets(items, basketsize=10):
    """generates balanced baskets from iterable, contiguous items"""
    num_items = len(items)
    num_baskets = max(1, num_items//basketsize)
    if num_items % basketsize and basketsize < num_items:
        num_baskets += 1
    step = num_items//num_baskets
    for i in range(0, num_items, step):
        yield items[i:i+step]


def chunks(l, n):
    """Yield successive n-sized chunks from l."""
    for i in range(0, len(l), n):
        yield l[i:i+n]


def create_parameter_range(domain):
    """returns the transposed array of the combined domain values"""
    L = [len(d) for d in domain]
    LS = np.prod(L)
    s = []
    e = 1
    for d in domain:
        LS = LS//len(d)
        s.append(np.repeat(d*LS, e))
        e = e*L[0]
        L = L[1:]
    return np.array(s).T


class Grid(object):
    """Parameter variation calculation"""
    def __init__(self, workdir,
                 magnetizingCurves=None, magnets=None, result_func=None): # tasktype='Task'):
        #self.tasktype = tasktype
        self.result_func = result_func
        self.femag = femagtools.Femag(workdir,
                                      magnetizingCurves=magnetizingCurves,
                                      magnets=magnets)
        self.stop = False  # rudimentary: gives the ability to stop a running parameter variation. thomas.maier/OSWALD
        self.reportdir=''
        """
        the "owner" of the Grid have to take care to terminate all running xfemag64 or wfemagw64
        processes after setting stop to True
        For example:
        
        def killFemagThreads():
            if sys.platform.startswith('linux'):
                os.system("kill $(ps aux | grep '[x]femag64' | awk '{print $2}')")
            else:
                os.system("taskkill /f /im wfemagw64.exe")
                
        thomas.maier/OSWALD
        """

    def set_report_directory(self, dirname):
        """saves the result files (BCH/BATCH) of every calculation 
        into this directory. Throws ValueError if directory is not empty or 
        FileNotFoundError if it does not exist.
        Args:
          dirname: name of report directory
        """
        if os.listdir(dirname):
            raise ValueError("directory {} is not empty".format(dirname))
        self.reportdir = dirname

        
    def setup_model(self, builder, model):
        """builds model in current workdir and returns its filenames"""
        # get and write mag curves
        mc_files = self.femag.copy_magnetizing_curves(model)
        
        filename = 'femag.fsl'
        logger.info("setup model in %s", self.femag.workdir)
        with open(os.path.join(self.femag.workdir, filename), 'w') as f:
            f.write('\n'.join(builder.create_model(model, self.femag.magnets) +
                              ['save_model("close")']))

        self.femag.run(filename, options=['-b'])
        model_files = [os.path.join(self.femag.workdir, m)
                       for m in mc_files] + [f
                                             for sublist in [
                                                     glob.glob(os.path.join(
                                                         self.femag.workdir,
                                                         model.name+e))
                                                     for e in (
                                                             '_*.poc',
                                                             '.nc', '.[IA]*7')]
                                             for f in sublist]
        
        logger.info("model %s created", model.name)
        return model_files
    
    def __call__(self, opt, pmMachine, operatingConditions,
                 engine, bchMapper=None):
        """calculate objective vars for all decision vars"""

        self.stop = False  # make sure the calculation will start. thomas.maier/OSWALD

        decision_vars = opt['decision_vars']
        objective_vars = opt.get('objective_vars', {})

        steps = [d.get('steps', 10) for d in decision_vars]
        logger.info('STEPS %s', str(steps))

        model = femagtools.model.MachineModel(pmMachine)
        # check if this model needs to be modified
        immutable_model = len([d for d in decision_vars
                               if hasattr(model,
                                          d['name'].split('.')[0])]) == 0
        operatingConditions['lfe'] = model.lfe
        operatingConditions['move_action'] = model.move_action
        operatingConditions['phi_start'] = 0.0
        operatingConditions['range_phi'] = 720/model.get('poles')
        operatingConditions.update(model.windings)
        fea = femagtools.model.FeaModel(operatingConditions)

        prob = femagtools.moproblem.FemagMoProblem(decision_vars,
                                                   objective_vars)

        job = engine.create_job(self.femag.workdir)
        builder = femagtools.fsl.Builder()

        # build x value array
        
        domain = [list(np.linspace(l, u, s))
                  for s, l, u in zip(steps, prob.lower, prob.upper)]

        par_range = create_parameter_range(domain)
        f = []
        p = 1
        calcid = 0
        logger.debug(par_range)

        if immutable_model:
            modelfiles = self.setup_model(builder, model)
            logger.info("Files %s", modelfiles)

        if hasattr(fea, 'poc'):
            fea.poc.pole_pitch = 2*360/model.get('poles')
            fea.pocfilename = fea.poc.filename()

        elapsedTime = 0
        self.bchmapper_data = []  # clear bch data
        # split x value (par_range) array in handy chunks:
        for population in baskets(par_range, opt['population_size']):
            if self.stop:  # try to return the results so far. thomas.maier/OSWALD
                logger.info(
                    'stopping grid execution... returning results so far...')
                try:
                    shape = [len(objective_vars)] + [len(d)
                                                     for d in reversed(domain)]
                    logger.debug("BEFORE: f shape %s --> %s",
                                 np.shape(np.array(f).T), shape)
                    completed = int(functools.reduce(
                        (lambda x, y: x * y), [len(z) for z in domain]))
                    logger.debug("need {} in total".format(completed))
                    remaining = completed - int(np.shape(np.array(f).T)[1])
                    values = int(np.shape(np.array(f).T)[0])
                    logger.debug("going to append {} None values".format(
                        remaining))
                    f += remaining * [values * [np.nan]]
                    shape = [len(objective_vars)] + [len(d)
                                                     for d in reversed(domain)]
                    logger.debug("AFTER: f shape %s --> %s",
                                 np.shape(np.array(f).T), shape)
                    objectives = np.reshape(np.array(f).T, shape)
                    r = dict(f=objectives.tolist(),
                             x=domain)
                    return r
                except:
                    return {}
                    pass

            logger.info('........ %d / %d results: %s',
                        p, len(par_range)//len(population)+1,
                        np.shape(f))
            job.cleanup()
            for k, x in enumerate(population):
                task = job.add_task(self.result_func)
                if immutable_model:
                    prob.prepare(x, fea)
                    for m in modelfiles:
                        task.add_file(m)
                    task.add_file(
                        'femag.fsl',
                        builder.create_open(model) +
                        builder.create_fe_losses(model) +
                        builder.create_analysis(fea) +
                        ['save_model("close")'])
                        
                else:
                    prob.prepare(x, [model, fea])
                    logger.info("prepare %s", x)
                    for mc in self.femag.copy_magnetizing_curves(
                            model,
                            task.directory):
                        task.add_file(mc)

                    task.add_file(
                        'femag.fsl',
                        builder.create_model(model, self.femag.magnets) +
                        builder.create_analysis(fea) +
                        ['save_model("close")'])
                if hasattr(fea, 'poc'):
                    task.add_file(fea.pocfilename,
                                  fea.poc.content())

            tstart = time.time()
            status = engine.submit()
            logger.info('Started %s', status)
            status = engine.join()
            tend = time.time()
            elapsedTime += (tend-tstart)
            logger.info("Elapsed time %d s Status %s",
                        (tend-tstart), status)
            for t in job.tasks:
                if t.status == 'C':
                    r = t.get_results()
                    # save result file if requested:
                    if self.reportdir:
                        repdir = os.path.join(self.reportdir,
                                              str(calcid))
                        os.makedirs(repdir)
                        try:
                            shutil.copy(glob.glob(os.path.join(
                                t.directory, r.filename)+'.B*CH')[0], repdir)
                        except (FileNotFoundError, AttributeError):
                            # must be a failure, copy all files
                            for ff in glob.glob(
                                    os.path.join(t.directory, '*')):
                                shutil.copy(ff, repdir)
                        calcid += 1
                    if isinstance(r, dict) and 'error' in r:
                        logger.warn("job %d failed: %s", k, r['error'])
                        f.append([float('nan')]*len(objective_vars))
                    else:
                        if bchMapper:
                            bchData = bchMapper(r)
                            self.addBchMapperData(bchData)
                            prob.setResult(bchData)
                        elif isinstance(r, dict):
                            prob.setResult(femagtools.getset.GetterSetter(r))
                        else:
                            prob.setResult(r)
                        f.append(prob.objfun([]))
                else:
                    f.append([float('nan')]*len(objective_vars))
            p += 1

        logger.info('Total elapsed time %d s ...... DONE', elapsedTime)

        shape = [len(objective_vars)] + [len(d) for d in reversed(domain)]
        logger.info("f shape %s --> %s", np.shape(np.array(f).T), shape)
        try:
            objectives = np.reshape(np.array(f).T, shape)
            if self.reportdir:
                self._write_report(decision_vars, objective_vars,
                                   objectives, domain)
            return dict(f=objectives.tolist(),
                        x=domain)
        except ValueError as v:
            logger.error(v)
            return dict(f=f, x=domain)
        
    def addBchMapperData(self, bchData):
        self.bchmapper_data.append(bchData)

    def getBchMapperData(self):
        return self.bchmapper_data

    def _write_report(self, decision_vars, objective_vars, objectives, domain):
        with open(os.path.join(self.reportdir, 'grid-report.csv'), 'w') as f:
            for line in get_report(decision_vars, objective_vars,
                                   objectives, domain):
                f.write(';'.join([str(v) for v in line]))
                f.write('\n')
                



# -*- coding: utf-8 -*-
"""
    femagtools.dxfsl.fslreader
    ~~~~~~~~~~~~~~~~~~~~~~~~~~

    renders geometry to fsl


"""
import numpy as np
from .shape import Shape
import logging
from .. import __version__
logger = logging.getLogger(__name__)


def agndst(da1, da2, Q, p, nodedist=1):
    """ build agndst from set of useful node angles:  
        4°, 2°, 1.5°, 1°, 0.75°, 0.5°, 0.25, 0.1, 0.05° """
    divisors = [1,2,4,8,10,16,20,40,80,160,200,400]
    dagset = [2*np.pi/Q/i for i in divisors]
    r = (da1 + da2)/4
    ag = abs(da1 - da2)/6
    i = max(np.argmin(np.abs(np.array(dagset) - np.arctan2(ag, r))), 1)
    nd = min(round(nodedist), i)
    try:
        logger.info("Num nodes/slot %g nodedist %g",
                    divisors[i-1], nodedist)
        if nodedist > 1:
            return dagset[i-nd]*r
        elif nodedist < 1 or i == 0:
            return dagset[i]*r
    except IndexError:
        pass
    return dagset[i-1]*r


class FslRenderer(object):
    """a model that can created by FSL"""

    def __init__(self, name):
        self.model = name
        self.mirror_axis = None
        self.fm_nlin = None
        self.shaft = None

    def mirror_nodechains(self, p0, p1):
        self.mirror_axis = np.array((p0, p1)).ravel().tolist()

    def material(self, p0):
        self.fm_nlin = p0

    def circle(self, center, radius, color='blue'):
        num = int(2*np.pi*radius)
        if num < 8:
            num = 8
        circle = [u'cx, cy = {}, {}'.format(center[0],
                                            center[1]),
                  u'nc_circle_m(cx,cy,{}, {})'.format(radius, num),
                  u'create_mesh_se(cx, cy)\n']
        self.content += circle

    def arc(self, startangle, endangle, center, radius, color='blue'):
        num = 0
#        if self.nodedist > 0:
#            s = startangle
#            d = endangle - s
#            n = int(d/(2*np.pi))
#            if n < 0:
#                n -= 1
#            d -= n*2*np.pi
#            num = int(radius*d/self.nodedist + 1)
#            if num < 3 and radius*d > self.nodedist:
#                num = 3

        p1 = (center[0] + radius*np.cos(startangle),
              center[1] + radius*np.sin(startangle))
        p2 = (center[0] + radius*np.cos(endangle),
              center[1] + radius*np.sin(endangle))
        self.content.append(
            u"nc_circle_m({}, {}, {}, {}, {}, {}, {})".format(
                p1[0], p1[1], p2[0], p2[1],
                center[0], center[1], num))

    def ellipse(self, center, width, height,
                rtheta, start_param, end_param, color='blue'):
        theta = np.arange(start_param, end_param, 0.2)
        xbegin = 0.5 * width * np.cos(theta)
        ybegin = 0.5 * height * np.sin(theta)
        theta = np.arange(end_param, end_param+0.1, 0.3)
        xend = 0.5 * width * np.cos(theta)
        yend = 0.5 * height * np.sin(theta)

        x = np.concatenate((xbegin, xend), axis=0)
        y = np.concatenate((ybegin, yend), axis=0)

        R = np.array([
            [np.cos(rtheta), -np.sin(rtheta)],
            [np.sin(rtheta),  np.cos(rtheta)]])
        x, y = np.dot(R, np.array([x, y]))
        x += center[0]
        y += center[1]
        nodes = list(zip(x, y))
        n0 = nodes[0]
        for n1 in nodes[1:]:
            self.content.append(
                u"nc_line({}, {}, {}, {}, {}) -- ellipse".format(
                    n0[0], n0[1], n1[0], n1[1], 0))
            n0 = n1

    def line(self, p1, p2, color='blue', e=None):
        num = 0
        # if self.nodedist > 0:
        #     l = la.norm(np.asarray(p1)-p2)
        #     num = int(l/self.nodedist + 1)
        if e is not None and e.has_attribute('auxline'):
            self.content.append(
                u"nc_line({}, {}, {}, {}, {}) -- auxiliary".format(
                    p1[0], p1[1], p2[0], p2[1], num))
        else:
            self.content.append(
                u"nc_line({}, {}, {}, {}, {})".format(
                    p1[0], p1[1], p2[0], p2[1], num))

    def sorted_elements(self, geom, inner=False):
        if inner:
            r = geom.max_radius
        else:
            r = geom.min_radius

        # return sorted list of distances from airgap and elements
        return sorted([(abs(r - np.linalg.norm(e.center_of_connection())), e)
                       for e in geom.elements(Shape)])

    def render(self, machine, inner=False, outer=False):
        '''create fsl statements with nodechains'''
        machine.set_alfa_and_corners()
        geom = machine.geom
        geom.split_lines_longer_than(geom.max_radius/4)
        self.content = []

        ndt_list = [(0.2, 1.5), (0.45, 2), (0.7, 3.0), (1.1, 3.0)]
        dist = geom.max_radius - geom.min_radius
        el_sorted = self.sorted_elements(geom, inner)

        n = 0
        for d, e in el_sorted:
            d_percent = d / dist
            if ndt_list[n][0] < d_percent:
                self.content.append(u'\nndt({}*agndst)\n'.
                                    format(ndt_list[n][1]))
                while len(ndt_list) and ndt_list[n][0] < d_percent:
                    n += 1
            e.render(self)

        self.content.append(u'\n')

        parts = int(machine.get_symmetry_part())
        self.content += [u'-- parts      = {}'.format(parts),
                         u'-- min_radius = {}'.format(geom.min_radius),
                         u'-- max_radius = {}'.format(geom.max_radius),
                         u'-- min_corner = {}, {}'.format(
                             geom.start_min_corner(0),
                             geom.start_min_corner(1)),
                         u'-- max_corner = {}, {}'.format(
                             geom.start_max_corner(0),
                             geom.start_max_corner(1)),
                         u'\n']
        if inner:
            self.content += [u'inner_da_start = {}'
                             .format(geom.dist_start_max_corner()),
                             u'inner_da_end = {}'
                             .format(geom.dist_end_max_corner())]
        if outer:
            self.content += [u'-- create air layer outside',
                             u'x0, y0 = {}, {}'.format(
                                 geom.start_max_corner(0),
                                 geom.start_max_corner(1)),
                             u'hair = 1.0',
                             u'r1 = {} + hair'.format(geom.max_radius),
                             u'r, phi = c2pr(x0, y0)',
                             u'x1, y1 = pr2c(r1, phi)']
            if geom.is_mirrored():
                self.content += [
                    u'x2, y2 = pr2c(r1, math.pi/m.tot_num_slot+phi/2)',
                    u'x3, y3 = pr2c(r, math.pi/m.tot_num_slot+phi/2)']
            else:
                self.content += [
                    u'x2, y2 = pr2c(r1, 2*math.pi/m.tot_num_slot+phi)',
                    u'x3, y3 = pr2c(r, 2*math.pi/m.tot_num_slot+phi)']

            self.content += [u'nc_line(x0, y0, x1, y1, 0)',
                             u'nc_circle_m(x1, y1, x2, y2, 0.0, 0.0, 0)',
                             u'nc_line(x2, y2, x3, y3, 0)']
            if geom.is_mirrored():
                self.content.append(
                    u'x0, y0 = pr2c(r1 - hair/2, math.pi/m.tot_num_slot/2+phi/4)')
            else:
                self.content.append(
                    u'x0, y0 = pr2c(r1 - hair/2, math.pi/m.tot_num_slot+phi/2)')

            self.content += [
                u'create_mesh_se(x0, y0)',
                u'\n',
                u'outer_da_start = {}'.format(
                    geom.dist_start_min_corner()),
                u'outer_da_end = {}'.format(
                    geom.dist_end_min_corner())]

        self.content.append(u'\n')
        self.content.append(u'x0_iron_tooth, y0_iron_tooth = 0.0, 0.0')
        self.content.append(u'x0_iron_yoke, y0_iron_yoke = 0.0, 0.0')
        self.content.append(u'x0_shaft, y0_shaft = 0.0, 0.0')
        self.content.append(u'\n')

        subregions = {}
        num_windings = 0
        num_magnets = 0
        for area in geom.list_of_areas():
            if area.number_of_elements() > 1:
                p = area.get_point_inside(geom)
                if p:
                    self.content.append(u"x0, y0 = {}, {}".format(p[0], p[1]))
                    # self.content.append(u"point(x0, y0, red, 4)")  # for debugging
                    self.content.append(u"create_mesh_se(x0, y0)")

                if area.is_winding():
                    if area.type not in subregions:
                        subregions[area.type] = 1
                    num_windings += 1
                    self.content.append(u'm.xcoil_{}, m.ycoil_{} = x0, y0'.
                                        format(num_windings, num_windings))

                elif area.is_magnet():
                    if area.type not in subregions:
                        subregions[area.type] = 1
                    num_magnets += 1
                    self.content.append(u'xmag[{}], ymag[{}] = x0, y0'.
                                        format(num_magnets, num_magnets))
                    self.content.append(u'mag_orient[{}] = {}'.
                                        format(num_magnets, area.phi))

                elif area.type > 0:
                    if area.type in subregions:
                        self.content.append(
                            u'add_to_subreg(x0, y0, "{}")'.
                            format(area.name()))
                    else:
                        subregions[area.type] = 1
                        self.content.append(
                            u'def_new_subreg(x0, y0, "{}", "{}")'.
                            format(area.name(), area.color()))
                    if area.is_stator_iron_yoke():
                        self.content.append(
                            u'x0_iron_yoke, y0_iron_yoke = x0, y0')
                    elif area.is_stator_iron_tooth():
                        self.content.append(
                            u'x0_iron_tooth, y0_iron_tooth = x0, y0')
                    elif area.is_rotor_iron():
                        self.content.append(
                            u'x0_iron_yoke, y0_iron_yoke = x0, y0')
                    elif area.is_shaft():
                        self.content.append(
                            u'x0_shaft, y0_shaft = x0, y0')

                self.content.append(u"\n")

        txt = [u"if x0_iron_yoke > 0.0 then",
               u"  if mcvkey_yoke ~= 'dummy' then",
               u'    def_mat_fm_nlin(x0_iron_yoke, y0_iron_yoke, "blue", mcvkey_yoke, 100)',
               u'  else',
               u'    def_mat_fm(x0_iron_yoke, y0_iron_yoke, ur, 100)',
               u'  end',
               u'end\n']
        self.content.append(u'\n'.join(txt))

        txt = [u"if x0_iron_tooth > 0.0 then",
               u"  if (mcvkey_teeth ~= 'dummy' and mcvkey_teeth ~= nil) then",
               u'    def_mat_fm_nlin(x0_iron_tooth, y0_iron_tooth, "blue", mcvkey_teeth, 100)',
               u'  else',
               u'    def_mat_fm(x0_iron_tooth, y0_iron_tooth, ur, 100)',
               u'  end',
               u'end\n']
        self.content.append(u'\n'.join(txt))

        txt = [u"if x0_shaft > 0.0 then",
               u"  if mcvkey_shaft ~= 'dummy' then",
               u'    def_mat_fm_nlin(x0_shaft, y0_shaft, "lightgrey", mcvkey_shaft, 100)',
               u'  else',
               u'    def_mat_fm(x0_shaft, y0_shaft, ur, 100)',
               u'  end',
               u'end\n']
        self.content.append(u'\n'.join(txt))

        if num_windings > 0:
            if geom.is_mirrored():
                self.content += [
                    u'    r, phi = c2pr(m.xcoil_1, m.ycoil_1)',
                    u'    m.xcoil_2, m.ycoil_2 = pr2c(r, {}*2.0 - phi)'.format(geom.alfa)]
            self.content.append(u'm.wdg_location  = 1.0 -- stator\n')

        if num_magnets > 0:
            self.content.append(u'mag_exists   = {}'.
                                format(num_magnets))
            if geom.is_mirrored():
                self.content.append(u'mag_mirrored = true')
            else:
                self.content.append(u'mag_mirrored = false')
            self.content.append(u'mag_alpha    = {}\n'
                                .format(geom.get_alfa()))

        if geom.is_mirrored():
            self.content.append(u'-- mirror')
            self.content.append(u'mirror_nodechains({}, {}, {}, {})\n'.format(
                geom.mirror_corners[1][0],   # max x1
                geom.mirror_corners[1][1],   # max y1
                geom.mirror_corners[0][0],   # min x2
                geom.mirror_corners[0][1]))  # min y2

        # angle after mirroring
        self.content.append(u'alfa = {}\n'.format(geom.get_alfa()))

        self.content += [u'-- rotate',
                         u'x1, y1 = {}, {}'.format(
                             geom.start_corners[0][0],
                             geom.start_corners[0][1])]  # min xy1
        if outer:
            self.content.append(u'x2, y2 = pr2c(r1, phi)')
        else:
            self.content.append(u'x2, y2 = {}, {}'.format(
                             geom.start_corners[1][0],
                             geom.start_corners[1][1]))  # max xy1

        if geom.is_mirrored():
            self.content.append(u'x3, y3 = pr2c(x2, alfa)')
            self.content.append(u'x4, y4 = pr2c(x1, alfa)')
        else:
            self.content += [u'x3, y3 = pr2c(r1, 2*math.pi/m.tot_num_slot+phi)',
                             u'x4, y4 = {}, {}'.format(
                                 geom.end_corners[0][0],
                                 geom.end_corners[0][1])]  # min xy4

        self.content.append(u'if {} > 1 then'.format(geom.num_variable()))
        if geom.corners_dont_match():
            txt = [u'  -- Warning: corners dont match',
                   u'  mirror_nodechains(x3, y3, x4, y4)',
                   u'  if {} > 2 then'.format(geom.num_variable()),
                   u'    my_alfa = alfa * 2',
                   u'    x3, y3 = pr2c(x2, my_alfa)',
                   u'    x4, y4 = pr2c(x1, my_alfa)',
                   u'    n = {} / 2'.format(geom.num_variable()),
                   u'    rotate_copy_nodechains(x1,y1,x2,y2,x3,y3,x4,y4,n-1)',
                   u'  end']
            self.content.append(u'\n'.join(txt))
        else:
            self.content.append(
                u'  rotate_copy_nodechains(x1,y1,x2,y2,x3,y3,x4,y4,{}-1)'
                .format(geom.num_variable()))
        self.content.append(u'end')

        self.content.append(u'alfa = {} * alfa'.format(geom.num_variable()))

        if self.fm_nlin:
            self.content.append(u'\nx0, y0 = {}, {}'. format(
                self.fm_nlin[0], self.fm_nlin[1]))

            mat = [u"if fm_nlin_mcvfile ~= 'dummy' then",
                   u"  if fm_nlin_mcvfile == 'air' then",
                   u"    def_mat_fm(x0,y0, 1.0, fm_nlin_rlen)",
                   u"  else",
                   u"    def_mat_fm_nlin(x0,y0, fm_nlin_colour, fm_nlin_mcvfile, fm_nlin_rlen)",
                   u"  end",
                   u"else",
                   u"  def_mat_fm(x0,y0, 1000.0, fm_nlin_rlen)",
                   u"end"]
            self.content.append(u'\n'.join(mat))

        if self.shaft:
            mat = [u'\nif shaft_mat==1 then',
                   u'  def_mat_fm_nlin(0.1,0.1,"lightgrey",fm_nlin_mcvfile_shft,fm_nlin_rlen)',
                   u'end']
            self.content.append(u'\n'.join(mat))

        if num_magnets:
            self.content += [
                u'-- pm magnets',
                u'if mag_exists > 0 then',
                u'  if( m.remanenc == nil ) then',
                u'     m.remanenc = 1.15',
                u'     m.relperm  = 1.05',
                u'  end',
                u'  if( m.orient == nil ) then',
                u'    m.orient = m.parallel',
                u'  end',
                u'  if( m.magncond == nil ) then',
                u'    m.magncond = 625000.0',
                u'  end',
                u'  if( m.rlen == nil ) then',
                u'    m.rlen = 100',
                u'  end',
                u'  for i=0, m.npols_gen-1 do',
                u'    for n=1, mag_exists do',
                u'      r, p = c2pr(xmag[n], ymag[n])',
                u'      x0, y0 = pr2c(r, i*mag_alpha+p)',
                u'      phi = (i*mag_alpha+mag_orient[n])*180/math.pi',
                u'      if ( i % 2 == 0 ) then',
                u'        phi = phi - 180',
                u'        color = "red"',
                u'      else',
                u'        color = "green"',
                u'      end',
                u'      if(m.mcvkey_magnet == nil) then',
                u'        def_mat_pm(x0, y0, color, m.remanenc, m.relperm,',
                u'                   phi, m.orient, m.magncond, m.rlen)',
                u'      else',
                u'        def_mat_pm_nlin(x0, y0, color, m.mcvkey_magnet,',
                u'                 phi, m.orient, m.magncond, m.rlen)',
                u'      end',
                u'      if mag_mirrored then',
                u'        x0, y0 = pr2c(r, (i+1)*mag_alpha-p)',
                u'        phi = ((i+1)*mag_alpha-mag_orient[n])*180/math.pi',
                u'        if ( i % 2 == 0 ) then',
                u'          phi = phi - 180',
                u'        end',
                u'        if(m.mcvkey_magnet == nil) then',
                u'          def_mat_pm(x0, y0, color, m.remanenc, m.relperm,',
                u'                       phi, m.orient, m.magncond, m.rlen)',
                u'        else',
                u'          def_mat_pm_nlin(x0, y0, color, m.mcvkey_magnet,',
                u'                   phi, m.orient, m.magncond, m.rlen)',
                u'        end',
                u'      end',
                u'    end',
                u'  end',
                u'end']

        return self.content

    def render_main(self,
                    motor,
                    m_inner, m_outer,
                    inner, outer,
                    params):
        '''create fsl statements for main file'''
        if not (m_inner and m_outer):
            logger.warning("ERROR: Rotor or Stator missing")
            return []

        num_layers = min(m_inner.num_of_layers() +
                         m_outer.num_of_layers(),
                         2)

        return [
            u'-- generated from DXF by femagtools {}'.format(__version__),
            u'exit_on_error = false',
            u'exit_on_end = false',
            u'verbosity = 2\n',
            u'new_model_force("{}","Test")'.format(self.model),
            u'global_unit(mm)',
            u'pickdist(0.001)',
            u'cosys(polar)\n',
            u'xmag = {}',
            u'ymag = {}',
            u'mag_orient = {}',
            u'mag_exists = 0',
            u'dy1 = {}'.format(params.get('dy1', 0.0)),
            u'da1 = {}'.format(params.get('da1', 0.0)),
            u'dy2 = {}'.format(params.get('dy2', 0.0)),
            u'da2 = {}'.format(params.get('da2', 0.0)),
            u'ag  = (da1 - da2)/2\n',
            u'm.tot_num_slot   = {}'.format(params.get('tot_num_slot', 0)),
            u'm.num_sl_gen     = {}'.format(params.get('num_sl_gen', 0)),
            u'm.num_poles      = {}'.format(params.get('num_poles', 0)),
            u'm.num_pol_pair   = m.num_poles/2',
            u'm.num_slots      = m.num_sl_gen',
            u'm.npols_gen      = m.num_poles * m.num_sl_gen / m.tot_num_slot',
            u'm.tot_num_sl     = m.tot_num_slot',
            u'm.fc_radius      = (da1+da2)/4',
            u'm.fc_radius1     = m.fc_radius',
            u'm.arm_length     = 1.0',
            u'pre_models("basic_modpar")\n',
            u'm.airgap         = 2*ag/3',
            u'm.nodedist       = 1.0',
            u'agndst           = {}'.format(params.get('agndst', 0.1)),
            u"mcvkey_teeth = 'dummy'",
            u"mcvkey_yoke = 'dummy'",
            u"mcvkey_shaft = 'dummy'",
            u"ur = 1000.0",
            u"ndt(agndst)"] + outer + [
                u"mcvkey_teeth = 'dummy'",
                u"mcvkey_yoke = 'dummy'",
                u"mcvkey_shaft = 'dummy'",
                u"ndt(agndst)"] + inner + [
                    u'-- airgap',
                    u'ndt(agndst)',
                    u'r1 = m.fc_radius - ag/6',
                    u'x1, y1 = pr2c(r1, alfa)',
                    u'n = math.floor(r1*alfa/agndst + 1.5)',
                    u'nc_circle_m(r1, 0, x1, y1, 0.0, 0.0, n)\n',
                    u'r2 = m.fc_radius + ag/6',
                    u'x2, y2 = pr2c(r2, alfa)',
                    u'nc_circle_m(r2, 0, x2, y2, 0.0, 0.0, n)\n',
                    u'if inner_da_start == nil then',
                    u'  inner_da_start = da2/2',
                    u'end',
                    u'x1, y1 = pr2c(inner_da_start, 0.0)',
                    u'nc_line(x1, y1, r1, 0.0, 0.0)\n',
                    u'if outer_da_start == nil then',
                    u'  outer_da_start = da1/2',
                    u'end',
                    u'x2, y2 = pr2c(outer_da_start, 0.0)',
                    u'nc_line(r2, 0.0, x2, y2, 0.0)\n',
                    u'if m.tot_num_slot > m.num_sl_gen then',
                    u'  x3, y3 = pr2c(inner_da_end, alfa)',
                    u'  x4, y4 = pr2c(r1, alfa)',
                    u'  nc_line(x3, y3, x4, y4, 0, 0)\n',
                    u'  x3, y3 = pr2c(outer_da_end, alfa)',
                    u'  x4, y4 = pr2c(r2, alfa)',
                    u'  nc_line(x3, y3, x4, y4, 0, 0)',
                    u'end\n',
                    u'x0, y0 = pr2c(r1-ag/6, alfa/2)',
                    u'create_mesh_se(x0, y0)',
                    u'x0, y0 = pr2c(r2+ag/6, alfa/2)',
                    u'create_mesh_se(x0, y0)\n',
                    u'connect_models()\n',
                    u'-- Gen_winding',
                    u'if m.xcoil_1 ~= nil then',
                    u'  m.num_phases      = 3',
                    u'  m.num_layers      = {}'.format(num_layers),
                    u'  m.num_wires       = 1',
                    u'  m.coil_span       = {}'.format(
                        2*num_layers*params.get(
                            'tot_num_slot', 1)//params.get('num_poles', 1)//3),
                    u'  m.current         = 0.0',
                    u'  m.mat_type        = 1.0 -- rotating',
                    u'  m.wind_type       = 1.0 -- winding & current',
                    u'  m.win_asym        = 1.0 -- sym',
                    u'  m.wdg_location    = 1.0 -- stator',
                    u'  m.curr_inp        = 0.0 -- const',
                    u'  m.dq_offset       = 0',
                    u'  pre_models("Gen_winding")',
                    u'  pre_models("gen_pocfile")',
                    u'end\n',
                    u'save_model(cont)\n']



# -*- coding: utf-8 -*-
"""
    femagtools.poc
    ~~~~~~~~~~~~~~

    Manage POC files



"""


class Poc:
    def __init__(self, arg, parameters=dict()):
        """initialize this object from a set of parameters or a file

        Args:
            arg filename or pole_pitch
        """
        for k in parameters.keys():
            self.__setattr__(k, parameters[k])

        if isinstance(arg, str):
            with open(arg) as f:
                self.readfile(f)
        else:
            self.pole_pitch = arg
            self.pocType = parameters.get('pocType', '')
            self.shape_current = parameters.get('shape_current', '')
            if not self.pocType:
                if self.shape_current:
                    self.pocType='Function'
                else:
                    self.shape_current = 'sin'
            
            self.key_winding = parameters.get('key_winding',
                                          [str(i+1) for i in range(3)])
            b = parameters.get('offset', 0)
            num_winding = len(self.key_winding)
            self.phi_voltage_winding = parameters.get('phi_voltage_winding',
                                                      [b+i*360/num_winding
                                                       for i in range(num_winding)])

    def __setattr__(self, name, val):
        self.__dict__[name] = val  # this will create the attribute name

    def filename(self):
        prefix = self.pocType
        if prefix == 'Function':
            prefix = self.shape_current
        return '{}_{}p.poc'.format(
            prefix, round(720/self.pole_pitch))
        
    def write(self, pocfilename):
        """create a new pocfile and write the data"""
        with open(pocfilename, mode='w') as pocfile:
            pocfile.write('\n'.join(self.content()))

    def content(self):
        """return list of lines"""
        # windings (num and keys)
        num_winding = len(self.key_winding)
        content = ["{0}".format(num_winding)]
        content.extend(["{0}".format(k)
                        for k in self.key_winding[:num_winding]])

        # phi voltage windings
        content.extend(["{0}".format(v)
                       for v in self.phi_voltage_winding[:num_winding]])

        # rest
        if self.pole_pitch:
            content.append("{0}".format(self.pole_pitch))
        if self.pocType in ['fun', 'har', 'hsp']:
            func_steps = len(self.func_current)
            content.append("{0}\n{1}".format(
                self.pocType, func_steps))
            if (self.pocType == 'fun' and
                num_winding*func_steps > len(self.func_current)):
                self.func_current = num_winding * self.func_current
                self.func_phi = num_winding * self.func_phi
                func_steps = num_winding*func_steps
            for i, val in enumerate(self.func_current[:func_steps]):
                if self.pocType == 'hsp':
                    content.append("{0}, {1}, {2}".format(
                        self.harmonic_id[i],
                        val, self.func_phi[i]))
                elif self.pocType == 'har':
                    content.append("{0}, {1}".format(
                        val, self.func_phi[i]))
                elif self.pocType == 'fun':
                    content.append("{0}, {1}".format(
                        self.func_phi[i], val))

        if self.pocType == 'Function':
            content.append("{0}".format(self.shape_current))

        if 'skew_angle' in self.__dict__:
            content.append("{0}".format(self.skew_angle))
        if 'num_skew_steps' in self.__dict__:
            content.append("{0}".format(self.num_skew_steps))
        content.append('')
        return content

    def readfile(self, pocfile):
        """read poc file"""
        self.num_winding = int(pocfile.readline())
        self.key_winding = []
        for i in range(self.num_winding):
            self.key_winding.append(int(pocfile.readline()))
        self.phi_voltage_winding = []
        for v in self.key_winding:
            self.phi_voltage_winding.append(float(pocfile.readline()))
        self.pole_pitch = float(pocfile.readline())
        self.pocType = pocfile.readline().strip()
        if self.pocType in ['fun', 'har', 'hsp']:
            self.func_current = []
            self.func_phi = []
            func_steps = int(pocfile.readline())
            if self.pocType == 'hsp':
                self.harmonic_id = []
            for i in range(func_steps):
                l = pocfile.readline().strip().split(',')
                if len(l) > 2:
                    self.harmonic_id.append(int(l[0]))
                    self.func_current.append(float(l[1]))
                    self.func_phi.append(float(l[2]))
                else:
                    self.func_current.append(float(l[0]))
                    self.func_phi.append(float(l[1]))
        else:
            self.shape_current=self.pocType
            self.pocType='Function'
        
        try:
            self.skew_angle=float(pocfile.readline())
            self.num_skew_steps=int(pocfile.readline())
        except ValueError:
            pass

    def getProps( self ):
        keys=['key_winding',
              'phi_voltage_winding',
              'pole_pitch',
              'pocType',
              'shape_current',
              'skew_angle',
              'num_skew_steps',
              'harmonic_id',
              'func_phi',
              'func_current']
        props={}
        for k in keys:
            if k in self.__dict__:
                props[k]=self.__dict__[k]
        return props

class HspPoc(Poc):
    def __init__(self, harm=[1], amp=[1], phi=[0]):
        Poc.__init__(self, 0, dict(
            pocType='hsp', harmonic_id=harm,
            func_current=amp, func_phi=phi))

        
def curr_har(x, n, A, phi ):
    "return fourier sum"
    import numpy as np
    if isinstance(A, list):
        return np.sum([curr_har(x, *nAp)
                       for nAp in zip(n, A, phi)], axis=0)
    return A*np.sin(n*x - phi)


def curr_fun(phi, A, phi_wdg ):
    "return curr"
    import numpy as np
    if np.asarray(A).shape == 1:
        return [[np.asarray(phi)+x, A] for x in phi_wdg] 
    return [phi, A]


if __name__ == "__main__":
    p = Poc('2p_sin.poc')
    print(p.getProps())

    p = Poc(60)
    print(p.getProps())

#    import numpy as np
#    import numpy.linalg as la
#    import matplotlib.pyplot as pl

#    A = [0.9866, 0.011181, 0.018624, 0.022322, 0.020109, 0.016582]
#    n = [1,11,13,14,16,17]
#    phi = [0, 225.801, 162.3245, 195.1087, 321.5113, -3.517]
#    x = np.linspace( 0, 2*np.pi, 40 )
#    y = curr( x, n, A,phi)
#    pl.plot( x, y )
#    pl.grid()
#    pl.show()



# -*- coding: utf-8 -*-
"""
    femagtools.condor
    ~~~~~~~~~~~~~~~~~

    Creating and managing condor jobs



"""
#
import os
import subprocess
import sys
import datetime
import re
import time
import json
import fnmatch
import logging
import femagtools.job

logger = logging.getLogger(__name__)


def getDateTime(date, time):
    """complete date with year: m/d --> y/m/d"""
    if date == "???":
        return ""
    today = datetime.date.today()
    m, d = date.split('/')
    fulldate = today.replace(month=int(m), day=int(d))
    if fulldate > datetime.date.today():  # greater than today => decrease year
        fulldate = fulldate.replace(year=(datetime.date.today().year-1))
    return "{0}T{1}:00".format(fulldate.isoformat(), time)


class CondorCluster(object):
    """manages condor cluster directories"""

    def __init__(self, userdir):
        self.basedir = userdir
        self.dataFile = 'cluster.json'

    def getAvailableClusters(self):
        results = []
        for fn in self.getClusterDirectories():
            try:
                with open(fn, 'r') as f:
                    vars = json.load(f)
                results.append(vars['clusterId'])
            except ValueError:
                continue
        return results

    def getClusterDirectories(self):
        results = []

        for base, dirs, files in os.walk(self.basedir):
            goodfiles = fnmatch.filter(files, self.dataFile)
            for f in goodfiles:
                results.append(os.path.join(base, f))
        return results

    def getClusterData(self, directory):
        for d in self.getClusterDirectories():
            logger.info("Query %s != %s", os.path.basename(d), directory)
            if os.path.basename(os.path.dirname(d)) == directory:
                logger.info("Found %s", d)
                with open(d, 'r') as f:
                    vars = json.load(f)
                    return vars

        logger.info("Not Found %s", d)
        return None

    def getClusterList(self, clusterIds=[]):
        results = []
        for fn in self.getClusterDirectories():
            try:
                with open(fn, 'r') as f:
                    vars = json.load(f)
            except ValueError:
                continue

            # tooltip
            s = "<h3>ClusterId {0}</h3><table>\n".format(vars['clusterId'])
            attrs = dict(variable='Variable', baseValue='BaseValue',
                         bounds=['StartValue','EndValue'],
                         nSteps='nSteps',unit='Unit', desc='Desc')
            for r in vars['optimize']['decision_vars']:
                for a in attrs:
                    if isinstance(attrs[a], str):
                        rs = re.sub('@.*?@', '', str(r[a]))
                        s += "<tr><td>{0}:</td><td>{1}</td></tr>".format(attrs[a], rs)
                    elif isinstance(attrs[a], list):
                        for i in range(len(r[a])):
                            rs = re.sub('@.*?@', '', str(r[a][i]))
                            s += "<tr><td>{0}:</td><td>{1}</td></tr>".format(attrs[a][i], rs)
            s += "</table>"

            # get infos from each cluster directory
            if 'clusterId' in vars:
                obj = dict(clusterId=vars['clusterId'],
                           directory=os.path.basename(os.path.dirname(fn)),
                           timestamp=time.strftime("%Y-%m-%dT%H:%M:%S",
                                                   time.localtime(os.path.getctime(fn))),
                           tooltip=s,
                           decision_vars=vars['optimize']['decision_vars'],
                           selected=(vars['clusterId'] in clusterIds))
                results.append(obj)
        return results

    def status(self, clusterId=None):
        if clusterId:
            return self.getClusterData(clusterId)
        else:
            return self.getClusterList()


class Engine(object):
    """manages calculation tasks in a HTCondor environment"""

    def __init__(self):
        self.job = None

    def create_job(self, workdir):
        self.job = femagtools.job.CondorJob(workdir)
        return self.job

    def status(self, userdir=None):
        status_cmd = 'condor_status'
        try:
            proc = subprocess.Popen([status_cmd],
                                    stdout=subprocess.PIPE,
                                    stderr=subprocess.PIPE)

            result = dict(machines=[], total=[])
            for l in proc.stdout:
                l = l.decode('utf-8').strip().split()
                if len(l) == 8 and \
                   (l[0].find(b'@') > 0 or str(l[5]).find('.') > 0):
                    result['machines'].append(dict(
                        name=l[0],
                        opsys=l[1],
                        arch=l[2],
                        state=l[3],
                        activity=l[4],
                        loadavg=float(l[5]),
                        mem=float(l[6]),
                        acttime=l[7]))
                elif len(l) == 8 and l[0] != b'Name':
                    result['total'].append(dict(oparch=l[0],
                                                total=int(l[1]),
                                                owner=int(l[2]),
                                                claimed=int(l[3]),
                                                unclaimed=int(l[4]),
                                                matched=int(l[5]),
                                                preempting=int(l[6]),
                                                backfill=int(l[7])))

            condorCl = CondorCluster(userdir)
            result['queue'] = self.queue(None)
            l = condorCl.getAvailableClusters()
            result['history'] = self.history(condorCl.getAvailableClusters())
            if userdir:
                result['cluster'] = condorCl.status()
            return result

        except ValueError as e:
            logger.error("Cannot invoke %s", status_cmd)
            e.args += status_cmd
            raise e

    def statusDir(self, userdir, directory):
        condorCl = CondorCluster(userdir)
        logger.info("%s, %s", userdir, directory)
        return condorCl.getClusterData(directory)

    def submit(self):
        submitfile = self.job.prepareDescription()
        cmdout = subprocess.check_output(["condor_submit", submitfile])
        self.clusterId = re.findall(r'\d+', cmdout.decode('utf-8'))[-1]
        logger.info('submit cluster %s directory %s total tasks %d',
                    self.clusterId, self.job.basedir, len(self.job.tasks))
        return self.clusterId

    def join(self):
        """wait for all tasks to be terminated and return status"""
        ret = []
        logger.info("CondorEngine.join")
        if self.clusterId:
            cmd = ["condor_q", "-json", self.clusterId]
            while True:
                time.sleep(2)
                cmdout = subprocess.check_output(cmd)
                if cmdout:
                    tasks = json.loads(cmdout.decode('utf-8'))
                if not cmdout or len(tasks) == 0:
                    break

            status = dict()
            cmdout = subprocess.check_output(
                ["condor_history", '-json', self.clusterId])
            hist = json.loads(cmdout.decode('utf-8'))
            for jobinfo in hist:
                taskid = jobinfo.get('ProcId', 0)
                status[taskid] = 'C' if jobinfo.get('JobStatus', 1) == 4 else \
                    'X' if jobinfo.get('JobStatus', 1) == 3 else 'E'
                logger.info('status %d: %s', taskid, status[taskid])
                if status[taskid] != 'X':
                    self.job.setExitStatus(taskid, status[taskid])

            for k in sorted(status.keys()):
                ret.append(status[k])
            logger.info('finished cluster %s', self.clusterId)
        else:
            logger.warn('no condor cluster')

        return ret

    def queue(self, clusterId):
        cmd = ["condor_q"]
        results = []
        if clusterId:
            cmd.append(str(clusterId))
        proc = subprocess.Popen(cmd,
                                stdout=subprocess.PIPE,
                                stderr=subprocess.PIPE)

        for l in proc.stdout:
            l = l.decode('utf-8').strip()
            if l.startswith('ID') or l.find('completed') > 0:
                continue
            props = l.split()
            if len(props) >= 9:
                s = props[4].split('+')
                t = s[1].split(':')
                c, p = props[0].split('.')
                results.append(dict(
                    id=props[0],
                    clusterId=c,
                    processId=p,
                    owner=props[1],
                    submitted=getDateTime(props[2], props[3]),
                    cpuUsage='{0:02}:{1}:{2}'.format(
                        int(24*s[0]+t[0]), t[1], t[2]),
                    status=props[5]))

        return results

    def history(self, clusterId):
        cmd = ["condor_history"]
        results = []
        if isinstance(clusterId, list):
            for clId in clusterId:
                results += self.history(str(clId))
            return results
        elif isinstance(clusterId, str):
            cmd.append(clusterId)

        logger.info("execute history cmd: " + str(cmd))
        proc = subprocess.Popen(cmd,
                                stdout=subprocess.PIPE,
                                stderr=subprocess.PIPE)

        for l in proc.stdout:
            l = l.decode('utf-8').strip()
            if l.startswith('ID') or l.find('completed') > 0:
                continue
            props = l.split()
            if len(props) >= 8:
                s = props[4].split('+')
                t = s[1].split(':')
                c, p = props[0].split('.')
                results.append(dict(
                    id=props[0],
                    clusterId=c,
                    processId=p,
                    owner=props[1],
                    submitted=getDateTime(props[2], props[3]),
                    cpuUsage='{0:02}:{1}:{2}'.format(
                        int(24*s[0] + t[0]), t[1], t[2]),
                    status=props[5],
                    completed=getDateTime(props[6], props[7])))
            else:
                sys.stderr.write("unknown condor_history line '{}'\n".format(
                    l))
        return results

    def terminate(self):
        logger.info("terminate Engine, clustId: %s", self.clusterId)
        proc = subprocess.Popen(['condor_rm', self.clusterId], shell=False)
        proc.wait()



"""
    femagtools.fsl
    ~~~~~~~~~~~~~~

    Creating FSL Scripts



"""
import logging
import mako
import mako.lookup
import os
import re
import sys
import math
from femagtools.dxfsl.converter import convert
from . import __version__
logger = logging.getLogger(__name__)


class FslBuilderError(Exception):
    pass


class Builder:
    def __init__(self):
        if getattr(sys, 'frozen', False) and hasattr(sys, '_MEIPASS'):
            # lookup up files in pyinstaller bundle
            logging.debug("Frozen!")
            dirs = [os.path.join(sys._MEIPASS, 'fsltemplates'),
                    os.path.join(os.getcwd(), '.')]
        else:
            dirs = [os.path.join(os.path.dirname(__file__), 'templates'),
                    os.path.join(os.getcwd(), '.')]
        self.lookup = mako.lookup.TemplateLookup(
            directories=dirs,
            disable_unicode=False, input_encoding='utf-8',
            output_encoding='utf-8',
            default_filters=['decode.utf8'])

        self.fsl_stator = False
        self.fsl_magnet = False

    def create_stator_model(self, model):
        mcv = ["mcvkey_yoke = '{}'"
               .format(model.stator.get('mcvkey_yoke', 'dummy')),
               "mcvkey_shaft = '{}'"
               .format(model.stator.get('mcvkey_shaft', 'dummy'))]
        if model.stator.get('mcvkey_teeth'):
            mcv.append("mcvkey_teeth = '{}'".format(
                model.stator['mcvkey_teeth']))

        return mcv + self.render_stator(model)

    def prepare_stator(self, model):
        templ = model.statortype()
        if templ == 'mshfile':
            import femagtools.gmsh

            g = femagtools.gmsh.Gmsh(model.stator['mshfile']['name'])
            phi = g.get_section_angles()
            model.stator['num_slots'] = round(math.pi/(phi[1]-phi[0]))
            r = g.get_section_radius()
            model.set_value('outer_diam', 2*r[1])
            model.set_value('bore_diam', 2*r[0])
            model.stator['stator_msh'] = dict(
                name=model.stator['mshfile']['name']
            )
            for sr in g.get_subregions():
                model.stator[sr] = g.get_location(sr)
            for k in ('yoke', 'teeth', 'air'):
                model.stator[k] = model.stator['mshfile'].get(k, [])
            wdg = model.stator['mshfile'].get('wdg', '')
            if wdg:
                model.stator['wdg'] = model.stator[wdg]
            del model.stator['mshfile']

            self.fsl_stator = True
            return

        if templ != 'dxffile':
            return

        logger.info("Conv stator from %s",
                    model.stator['dxffile']['name'])
        params = {}
        params['split'] = model.stator[templ].get('split', False)
        params['show_plots'] = model.stator[templ].get('plot', False)
        params['write_fsl'] = True
        params['airgap'] = -1.0
        pos = 'in' if model.external_rotor else 'out'
        params['part'] = ('stator', pos)
        conv = convert(model.stator['dxffile']['name'], **params)

        model.stator['num_slots'] = conv.get('tot_num_slot')
        self.set_diameter_parameter(model, conv)
        if model.get('dy1'):
            model.set_value('outer_diam', model.get('dy1'))
            model.set_value('bore_diam', model.get('da1'))

        model.stator['dxf'] = dict(fsl=conv['fsl'])
        self.fsl_stator = True
        del model.stator[templ]

    def set_diameter_parameter(self, model, conv):
        for v in ('dy1', 'da1', 'da2', 'dy2'):
            try:
                model.set_value(v, conv[v])
            except KeyError:
                pass

    def prepare_diameter(self, model):
        dy1 = model.get('dy1', 0.0)
        if dy1:
            model.set_value('outer_diam', dy1 * 1e-3)
        da1 = model.get('da1', 0.0)
        if da1:
            model.set_value('bore_diam', da1 * 1e-3)
        dy2 = model.get('dy2', 0.0)
        if dy2:
            model.set_value('inner_diam', dy2 * 1e-3)
        da2 = model.get('da2', 0.0)
        if da1 and da2:
            model.set_value('airgap', abs(da1 - da2)/2/1e3)

    def render_stator(self, model):
        templ = model.statortype()
        if templ == 'dxf':
            return [
                'agndst = {}'.format(model.agndst*1e3),
                'ndt(agndst)'
            ] + model.stator['dxf']['fsl']
        if templ == 'statorFsl':
            #  obsolete
            if 'parameter' in model.stator['statorFsl']:
                return self.render_template(
                    model.stator['statorFsl']['content_template'],
                    model.stator['statorFsl']['parameter'])
            if isinstance(model.stator['statorFsl']
                          ['content_template'], str):
                with open(model.stator['statorFsl']
                          ['content_template']) as f:
                    templ = [l.strip() for l in f.readlines()]
            else:
                templ = model.stator['statorFsl']['content_template']
            return self.render_template(
                '\n'.join(templ),
                model.stator['statorFsl'])

        statmodel = model.stator.copy()
        statmodel.update(model.stator[templ])
        statmodel.update({
            'zeroangle': model.stator.get('zeroangle', 0),
            'rlength': model.stator.get('rlength', 1),
            'num_layers': model.windings.get('num_layers', 1)}
        )

        fslcode = self.__render(statmodel, templ, stator=True)
        if fslcode:
            if self.fsl_stator:
                return (['agndst = {}'.format(model.agndst*1e3),
                         'alfa = 2*math.pi*m.num_sl_gen/m.tot_num_slot',
                         'num_agnodes = math.floor(m.fc_radius*alfa/agndst + 1.5)'] +
                        fslcode)
            return (fslcode +
                    ['post_models("nodedistance", "ndst" )',
                     'agndst=ndst[1]*1e3'])

        logger.error('File {}.mako not found'.format(templ))
        return []

    def create_magnet_model(self, model):
        mcv = ["mcvkey_yoke  = '{}'"
               .format(model.magnet.get('mcvkey_yoke', 'dummy')),
               "mcvkey_shaft = '{}'"
               .format(model.magnet.get('mcvkey_shaft', 'dummy'))]

        if 'magnetFsl' in model.magnet:
            #  obsolete
            if 'parameter' in model.magnet['magnetFsl']:
                return mcv + self.render_template(
                    model.magnet['magnetFsl']['content_template'],
                    model.magnet['magnetFsl']['parameter'])
            if isinstance(model.magnet['magnetFsl']
                          ['content_template'], str):
                with open(model.magnet['magnetFsl']
                          ['content_template']) as f:
                    templ = [l.strip() for l in f.readlines()]
            else:
                templ = model.magnet['magnetFsl']['content_template']

            return mcv + self.render_template(
                '\n'.join(templ),
                model.magnet['magnetFsl'])

        templ = model.magnettype()
        magmodel = model.magnet.copy()
        magmodel.update(model.magnet[templ])
        magmodel['mcvkey_magnet'] = model.get_mcvkey_magnet()
        if templ == 'dxf':
            return mcv + [
                u'xmag = {}',
                u'ymag = {}',
                u'mag_orient = {}',
                u'ndt(agndst)'] + model.magnet['dxf']['fsl']

        return mcv + self.render_rotor(magmodel, templ)

    def prepare_rotor(self, model):
        templ = model.magnettype()
        if templ == 'mshfile':
            import femagtools.gmsh
            g = femagtools.gmsh.Gmsh(model.magnet['mshfile']['name'])
            r = g.get_section_radius()
            phi = g.get_section_angles()
            p = round(math.pi/(phi[1]-phi[0]))
            model.set_value('poles', p)
            model.set_value('inner_diam', 2*r[0])
            ag = (model.get('bore_diam') - 2*r[1])/2
            model.set_value('airgap', ag)
            model.magnet['rotor_msh'] = dict(
                name=model.magnet['mshfile']['name']
            )
            for sr in g.get_subregions():
                model.magnet[sr] = g.get_location(sr)

            for k in ('yoke', 'air'):
                model.magnet[k] = model.magnet['mshfile'].get(k, [])
            model.magnet['mag'] = dict(
                sreg=model.magnet['mshfile'].get('mag', []),
                axis=[g.get_axis_angle(s)
                      for s in model.magnet['mshfile'].get('mag', [])]
            )
            del model.magnet['mshfile']
            return

        if templ != 'dxffile':
            return

        params = {}
        params['split'] = model.magnet[templ].get('split', False)
        params['show_plots'] = model.magnet[templ].get('plot', False)
        params['write_fsl'] = True
        params['airgap'] = -1.0
        pos = 'out' if model.external_rotor else 'in'
        params['part'] = ('rotor', pos)
        logger.info("Conv rotor from %s", templ + '.dxf')
        conv = convert(model.magnet[templ]['name'], **params)
        model.set_value('poles', int(conv.get('num_poles')))
        self.set_diameter_parameter(model, conv)
        if model.get('da2'):
            logger.info('da2 %f',  model.get('da2')/1e3)
            ag = (model.get('bore_diam') - model.get('da2')/1e3)/2
            model.set_value('airgap', ag)

        model.magnet['dxf'] = dict(fsl=conv['fsl'])
        self.fsl_magnet = True
        del model.magnet[templ]

    def render_rotor(self, magmodel, templ):
        fslcode = self.__render(magmodel, templ, magnet=True)
        if fslcode:
            return fslcode

        logger.error('File {}.fsl not found'.format(templ))
        return []

    def create_connect_models(self, model):
        """return connect_model if rotating machine and incomplete model
        (Note: femag bug with connect model)"
        """
        if (model.get('move_action') == 0 and (model.connect_full or
            model.stator['num_slots'] > model.stator['num_slots_gen'])):
            return ['pre_models("connect_models")\n']
        return []

    def create_open(self, model):
        return (['-- created by femagtools {}'.format(__version__), ''] +
                self.__render(model, 'open') +
                self.__render(model, 'basic_modpar'))

    def set_modpar(self, model):
        return self.__render(model, 'basic_modpar')

    def create_new_model(self, model):
        return (['-- created by femagtools {}'.format(__version__), ''] +
                self.__render(model, 'new_model'))

    def create_cu_losses(self, model):
        return self.__render(model.windings, 'cu_losses')

    def create_fe_losses(self, model):
        if any(model.get(k, 0) for k in ('ffactor', 'cw', 'ch', 'hyscoef',
                                         'edycof', 'indcof', 'fillfact',
                                         'basfreq', 'basind')):
            return self.__render(model, 'FE-losses')
        return []

    def create_gen_winding(self, model):
        if 'leak_dist_wind' in model.windings:
            return self.__render(model, 'gen_winding') + \
                self.__render(model.windings['leak_dist_wind'],
                              'leak_dist_wind')
        elif 'leak_evol_wind' in model.windings:
            return self.__render(model, 'gen_winding') + \
                self.__render(model.windings['leak_evol_wind'],
                              'leak_evol_wind')
        elif 'leak_tooth_wind' in model.windings:
            return self.__render(model, 'gen_winding') + \
                self.__render(model.windings['leak_tooth_wind'],
                              'leak_tooth_wind')
        return self.__render(model, 'gen_winding')

    def prepare_model_with_dxf(self, model):
        dxfname = model.dxffile.get('name', None)
        if not dxfname:
            logger.error('Name of dxf-file expected')
            return []

        if dxfname.split('.')[-1] not in ('dxf'):
            dxfname += '.dxf'
        if not os.path.isfile(dxfname):
            logger.error('File {} not found'.format(dxfname))
            return []

        params = {}
        params['split'] = model.dxffile.get('split', False)
        params['show_plots'] = model.dxffile.get('plot', False)
        params['write_fsl'] = True
        params['airgap'] = model.dxffile.get('airgap', 0.0)
        params['nodedist'] = model.dxffile.get('nodedist', 1)

        conv = convert(dxfname, **params)

        model.set_value('poles', conv.get('num_poles'))
        model.set_value('outer_diam', conv.get('dy1') * 1e-3)
        model.set_value('bore_diam', conv.get('da1') * 1e-3)
        model.set_value('inner_diam', conv.get('dy2') * 1e-3)
        model.set_value('airgap', (conv.get('da1') - conv.get('da2'))/2/1e3)
        model.set_value('agndst', conv.get('agndst')*1e-3)

        if not hasattr(model, 'stator'):
            setattr(model, 'stator', {})
        model.stator['num_slots'] = conv.get('tot_num_slot')
        model.stator['num_slots_gen'] = conv.get('num_sl_gen')
        if 'fsl_stator' in conv:
            self.fsl_stator = True
            model.stator['dxf'] = dict(fsl=conv['fsl_stator'])
        if not hasattr(model, 'magnet'):
            setattr(model, 'magnet', {})
        if 'fsl_magnet' in conv:
            self.fsl_magnet = True
            model.magnet['dxf'] = dict(fsl=conv['fsl_magnet'])

    def create_model(self, model, magnets=[]):
        if model.is_complete():
            if model.is_dxffile():
                self.prepare_model_with_dxf(model)
            else:
                self.prepare_stator(model)
                self.prepare_rotor(model)
                self.prepare_diameter(model)
                if self.fsl_stator:
                    from femagtools.dxfsl.fslrenderer import agndst
                    ag = model.get('airgap')
                    model.set_value(
                        'agndst',
                        agndst(model.get('bore_diam'),
                               model.get('bore_diam') - 2*ag,
                               model.stator.get('num_slots'),
                               model.get('poles'),
                               model.stator.get('nodedist') or 1.0))

                model.set_num_slots_gen()

            material = model.magnet.get('material', 0)
            magnetMat = {}
            if magnets and material:
                magnetMat = magnets.find(material)
                if not magnetMat:
                    raise FslBuilderError(
                        'magnet material {} not found'.format(
                            material))
                try:
                    magnetMat['magntemp'] = model.magn_temp
                except AttributeError:
                    magnetMat['magntemp'] = 20

            return (self.create_new_model(model) +
                    self.create_cu_losses(model) +
                    self.create_fe_losses(model) +
                    self.create_stator_model(model) +
                    self.create_gen_winding(model) +
                    self.create_magnet(model, magnetMat) +
                    self.create_magnet_model(model) +
                    self.mesh_airgap(model) +
                    self.create_connect_models(model))

        return self.open_model(model)

    def open_model(self, model):
        return self.create_open(model)

    def load_model(self, model):
        return self.__render(model, 'open')

    def create_magnet(self, model, magnetMat=None):
        if magnetMat:
            logger.info("Setting magnet properties %s", magnetMat['name'])
            if 'rlen' in model.magnet:
                magnetMat['rlen'] = model.magnet['rlen']
            return self.__render(magnetMat, 'magnet-data')
        return ['m.remanenc       = {}'
                .format(model.magnet.get('remanenc', 1.2)),
                'm.relperm        = {}'
                .format(model.magnet.get('relperm', 1.05)),
                'm.rlen           = {}'
                .format(model.magnet.get('rlen', 100)),
                '']

    def create_analysis(self, model):
        airgap_induc = (self.create_airgap_induc()
                        if model.get('airgap_induc', 0) else [])
        felosses = self.create_fe_losses(model)
        fslcalc = (self.__render(model, model.get('calculationMode')) +
                   airgap_induc)

        if model.get('calculationMode') in ('cogg_calc',
                                            'ld_lq_fast',
                                            'pm_sym_loss',
                                            'torq_calc',
                                            'psd_psq_fast'):
            return felosses + fslcalc

        return (felosses + fslcalc +
                self.__render(model, 'plots'))

    def create_shortcircuit(self, model):
        return self.__render(model, 'shortcircuit')

    def create_airgap_induc(self):
            return self.__render(dict(), 'airgapinduc')

    def create_colorgrad(self, model):
            return self.__render(model, 'colorgrad')

    def mesh_airgap(self, model):
        if self.fsl_stator and self.fsl_magnet:
            return self.__render(model, 'mesh-airgap')
        else:
            return []

    def create(self, model, fea, magnets=None):
        "create model and analysis function"
        try:
            fea['lfe'] = model.get('lfe')
        except AttributeError:
            pass
        try:
            fea['move_action'] = model.get('move_action')
        except AttributeError:
            pass
        try:
            fea.update(model.windings)
        except AttributeError:
            pass

        if model.is_complete():
            logger.info("create new model and simulation")
            fslmodel = self.create_model(model, magnets)
            if 'num_poles' in model.windings:
                num_poles = model.windings['num_poles']
            else:
                num_poles = model.get('poles')
            if 'poc' in fea:
                poc = fea['poc']
                poc.pole_pitch = 2*360/num_poles
                fea['pocfilename'] = poc.filename()
            else:
                fea['pocfilename'] = (model.get('name') +
                                      '_' + str(num_poles) +
                                      'p.poc')

            if 'phi_start' not in fea:
                fea['phi_start'] = 0.0
            if 'range_phi' not in fea:
                fea['range_phi'] = 720/model.get('poles')

            return (fslmodel + self.create_analysis(fea) +
                    ['save_model("close")'])

        logger.info("create open model and simulation")
        return (self.open_model(model) +
                self.create_analysis(fea) +
                ['save_model("close")'])

    def __render(self, model, templ, stator=False, magnet=False):
        if templ.split('.')[-1] in ('fsl', 'mako'):
            try:
                template = self.lookup.get_template(templ)
                logger.info('use file {}'.format(templ))
                return template.render_unicode(model=model).split('\n')
            except mako.exceptions.TopLevelLookupException as ex:
                logger.error('File {} not found'.format(templ))
                sys.exit(1)

        try:
            template = self.lookup.get_template(templ+".mako")
            logger.debug('use template {}.mako'.format(templ))
        except mako.exceptions.TopLevelLookupException as ex:
            template = self.lookup.get_template(templ+".fsl")
            logger.debug('use FSL {}.fsl'.format(templ))
            if stator:
                self.fsl_stator = True
            if magnet:
                self.fsl_magnet = True

        return template.render_unicode(model=model).split('\n')

    def render_template(self, content_template, parameters):
        template = mako.template.Template(content_template)
        obj = {}
        if isinstance(parameters, list):
            for p in parameters:
                obj[p['key']] = p['value']
            return template.render_unicode(
                par=obj).split('\n')
        return template.render_unicode(
            model=parameters).split('\n')

    def read(self, fslfile):
        """extracts parameters from content and creates template"""
        parpat = re.compile(
            r'''([\w\.]+)\s*= # key
            \s*(([+-]?\d+((?:\.\d+)?(?:[eE][+-]\d+)?))|
            (['"][^'"]*['"]))\s*-- # value
            \s*(.+)$''', re.X)
        content = []
        content_template = []
        parameter = []
        for line in fslfile:
            # add line to content
            content.append(line.strip())
            p = parpat.findall(line)
            if p:
                par = dict(key=p[0][0],
                           value=p[0][1].strip(),
                           comment=p[0][-1])
                parameter.append(par)
                content_template.append("{0} = {1} -- {2}".format(
                    par['key'],
                    '${par[\''+par['key']+'\']}',
                    par['comment']))
                continue

            # nothing do do => insert this line in template
            content_template.append(line.strip())

        return dict(
            content='\n'.join(content),
            type='fsl',
            content_template='\n'.join(content_template),
            parameter=parameter)
